
#ifndef BFM_LINALG_H
#define BFM_LINALG_H

/*  2007
 *  Calls PAB's assembler routines to give an implementation
 *  of the dwf dslash. Aim is to give very high performance
 *  in a reasonably portable/retargettable library.
 */
#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>

#define ALIGNIT(A) (double *)( (((uint64_t)A) + 31)& (~0x1FUL) );

template <class Float>
double bfmbase<Float>::axpby_norm(Fermion_t r, Fermion_t x, Fermion_t y,double a,double b)
{
  double nrm_b[4*4] = {0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
  double aa_b[8];
  double bb_b[8];
  double *nrm = ALIGNIT(nrm_b);
  double *aa  = ALIGNIT(aa_b);
  double *bb  = ALIGNIT(bb_b);

  int me,thrlen,throff;
  thread_work(2*cbLs*simd_cbvol*nsimd,me,thrlen,throff);// units of 6 Floats

  Float *xx = ((Float * )x) + throff*12;
  Float *yy = ((Float * )y) + throff*12;
  Float *rr = ((Float * )r) + throff*12;

  aa[0] = aa[1] = aa[2] = aa[3] = a;
  bb[0] = bb[1] = bb[2] = bb[3] = b;

  // new version simd enabled
  thrlen = thrlen/nsimd;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpby_norm((double *)rr,aa,bb,(double *)xx,(double *)yy,thrlen,nrm);
  } else { 
    vmx_vaxpby_norm_s((float *)rr,aa,bb,(float *)xx,(float *)yy,thrlen,nrm);
  }
  double dot = nrm[0];
  double mydot = dot;
  
  thread_sum(dot,me);

  if ( reproduce ) {
    repro(mydot,me);
    if ( reproduce_checksum ){
      uint64_t sum=checksum(r);
      if ( me == 0 ) repro(sum);
    }
    thread_barrier();
    if ( me == 0 ) repro(dot,nthread);
  }

  comm_gsum(dot);
  return dot;
}
template <class Float>
double bfmbase<Float>::axpy_norm(Fermion_t r, Fermion_t x, Fermion_t y,double a)
{
  double nrm_b[4*3+4] = {0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
  double aa_b[4+4]; // 64 bytes
  double *aa  = ALIGNIT(aa_b);
  double *nrm = ALIGNIT(nrm_b);

  int me,thrlen,throff;
  thread_work(2*cbLs*simd_cbvol*nsimd,me,thrlen,throff);// units of 6 Floats

  thrlen = thrlen/nsimd;

  Float *xx = ((Float * )x) + throff*12;
  Float *yy = ((Float * )y) + throff*12;
  Float *rr = ((Float * )r) + throff*12;

  aa[0] = aa[1] = aa[2] = aa[3] = a;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpy_norm((double *)rr,aa,(double *)xx,(double *)yy,thrlen,nrm);
  } else { 
    vmx_vaxpy_norm_s((float *)rr,aa,(float *)xx,(float *)yy,thrlen,nrm);
  }
  double dot = nrm[0];
  double mydot = dot;

  thread_sum(dot,me);

  if ( reproduce ) { 
    repro(mydot,me);
    if ( reproduce_checksum  ){
      uint64_t sum=checksum(r);
      if ( me == 0 ) repro(sum);
    }
    thread_barrier();
    if ( me == 0 ) repro(dot,nthread);
  }

  comm_gsum(dot);

  return dot;
}

template <class Float>
void bfmbase<Float>::axpibg5x(Fermion_t r, Fermion_t x, double a,double b)
{
  //r = a x + i b g5 x
  double bb_b[4+4];
  double aa_b[4+4]; // 64 bytes

  double *aa = ALIGNIT(aa_b);
  double *bb = ALIGNIT(bb_b);

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol,me,thrlen,throff);// units of 24 Floats
  Float *xx = ((Float * )x) + throff*24*nsimd;
  Float *rr = ((Float * )r) + throff*24*nsimd;

  aa[0] = aa[1] = aa[2] = aa[3] = a;
  bb[0] = bb[1] = bb[2] = bb[3] = b;
  
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_tmass((double *)rr,
               (double *)aa,
               (double *)bb,
               (double *)xx,
               thrlen);
  } else { 
    vmx_tmass_s((float *)rr,
		(double *)aa,
		(double *)bb,
		(float *)xx,
                thrlen);
  }

  thread_barrier();
}

template <class Float>
void bfmbase<Float>::merge(Float *r, Float *lo, Float *hi, int nsite)
{
  // nsite should be simd_nbound[mu]*cbLs
  int me,thrlen,throff;
  thread_work(nsite,me,thrlen,throff);// units of  Floats

  Float *llo = ((Float * )lo) + throff*6*nsimd;
  Float *hhi = ((Float * )hi) + throff*6*nsimd;
  Float *rr  = ((Float * )r)  + throff*12*nsimd;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_merge ((double *)rr,
               (double *)llo,
               (double *)hhi,
               thrlen);
  } else { 
    vmx_merge_s((float *)rr,
		(float *)llo,
		(float *)hhi,
                thrlen);
  }

  thread_barrier();
}

template <class Float>
void bfmbase<Float>::axpby(Fermion_t r, Fermion_t x, Fermion_t y,double a,double b)
{
  // write_vaxpy 2 way unrolled operating on 3 colors * complex * nsimd
  double bb_b[4+4];
  double aa_b[4+4]; // 64 bytes

  double *aa = ALIGNIT(aa_b);
  double *bb = ALIGNIT(bb_b);

  int me,thrlen,throff;
  thread_work(2*cbLs*simd_cbvol,me,thrlen,throff);// units of 6 Floats

  Float *xx = ((Float * )x) + throff*12*nsimd;
  Float *yy = ((Float * )y) + throff*12*nsimd;
  Float *rr = ((Float * )r) + throff*12*nsimd;

  aa[0] = aa[1] = aa[2] = aa[3] = a;
  bb[0] = bb[1] = bb[2] = bb[3] = b;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpby((double *)rr,
               (double *)aa,
               (double *)bb,
               (double *)xx,
               (double *)yy,thrlen);
  } else { 
    vmx_vaxpby_s((float *)rr,
                      (double *)aa,
                      (double *)bb,
                      (float *)xx,
                      (float *)yy,thrlen);
  }

  thread_barrier();
}




template <class Float>
double bfmbase<Float>::cg_psi_p(Fermion_t psi, Fermion_t p, Fermion_t r,Fermion_t mmp,double a,double b)
{
  // 2 way unrolled operating on 3 colors * complex * nsimd
  double bb_b[4+4];
  double aa_b[4+4]; // 64 bytes
  double nrm_b[4*4] = {0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};

  double *aa = ALIGNIT(aa_b);
  double *bb = ALIGNIT(bb_b);
  double *nrm = ALIGNIT(nrm_b);

  int me,thrlen,throff;
  thread_work(2*cbLs*simd_cbvol,me,thrlen,throff);// units of 6 Floats

  Float *ppsi = ((Float * )psi) + throff*12*nsimd;
  Float *pp   = ((Float * )p) + throff*12*nsimd;
  Float *rr   = ((Float * )r) + throff*12*nsimd;
  Float *mmmp = ((Float * )mmp) + throff*12*nsimd;

  aa[0] = aa[1] = aa[2] = aa[3] = a;
  bb[0] = bb[1] = bb[2] = bb[3] = b;
  integer iargs[10];
  iargs[0] = (integer)aa;
  iargs[1] = (integer)bb;
  iargs[2] = (integer)ppsi;
  iargs[3] = (integer)pp;
  iargs[4] = (integer)rr;
  iargs[5] = (integer)mmmp;
  iargs[6] = (integer)thrlen;
  iargs[7] = (integer)nrm;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_cg(iargs);
  } else { 
    vmx_cg_s(iargs);
  }
  double dot = nrm[0]+nrm[1]+nrm[2]+nrm[3];
  thread_sum(dot,me);
  comm_gsum (dot);

  return(dot);
}

template <class Float>
void bfmbase<Float>::axpy(Fermion_t r, Fermion_t x, Fermion_t y,double a)
{
  int me,thrlen,throff;

  double aa_b[4+4]; // 64 bytes
  double *aa = ALIGNIT(aa_b);

  thread_work(2*cbLs*simd_cbvol,me,thrlen,throff);// units of 6 complex

  Float *xx = ((Float * )x) + throff*12*nsimd;
  Float *yy = ((Float * )y) + throff*12*nsimd;
  Float *rr = ((Float * )r) + throff*12*nsimd;

  for(int i=0;i<4;i++) aa[i] = a;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpy((double *)rr,(double *)aa,(double *)xx,(double *)yy,thrlen);
  } else { 
    vmx_vaxpy_s((float *)rr,(double *)aa,(float *)xx,(float *)yy,thrlen);
  } 

  thread_barrier();

  return;
}

/* 
 * Performance non-critical
 */
template <class Float>
void bfmbase<Float>::scale(Fermion_t x_t, double a)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  Float nrm = 0.0;

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=0;s<thrlen;s++){

    xx = &x[(s+throff)*nspinco*2 ];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + spinco*2;
      xx[idx] = a *xx[idx];
    }}

  }  
  
  thread_barrier();
}


template <class Float>
void bfmbase<Float>::scale(Fermion_t x_t, double ar, double ai)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  Float nrm = 0.0;

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=0;s<thrlen;s++){

    xx = &x[(s+throff)*nspinco*2 ];

    for(int spinco=0;spinco<nspinco;spinco++){
      int idx = spinco*2;
      double xr = xx[idx];
      double xi = xx[idx+1];
      xx[idx]   = ar*xr - ai*xi;
      xx[idx+1] = ai*xr + ar*xi;
    }

  }  
  thread_barrier();
}

template <class Float>
void bfmbase<Float>::copy(Fermion_t x_t, Fermion_t copyme)
{
  axpy(x_t,copyme,copyme,0.0);
}

 
template <class Float>
void bfmbase<Float>::set_zero(Fermion_t x_t)
{
  this->fill(x_t,0.0);
}

#ifdef BFM_QDP_BINDING
template <class Float>
complex<double> bfmbase<Float>::inner(Fermion_t x_t,Fermion_t y_t)
{
  int nspinco = 12;

  Float *x = (Float *) x_t;
  Float *y = (Float *) y_t;
  Float *xx;
  Float *yy;
  double rr = 0.0;
  double ii = 0.0;


  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=throff;s<throff+thrlen;s++){

    xx = &x[s*nspinco*2];
    yy = &y[s*nspinco*2];

    for(int spinco=0;spinco<nspinco;spinco++){
      int idx = spinco*2;
      double xr = xx[idx];
      double xi = xx[idx+1];
      double yr = yy[idx];
      double yi = yy[idx+1];
      rr = rr + xr*yr+xi*yi;
      ii = ii - xi*yr+xr*yi;
    }
  }

  thread_sum(rr,me);
  comm_gsum(rr);

  thread_sum(ii,me);
  comm_gsum(ii);

  complex<double> dot(rr,ii);
  return dot;
}

template <class Float>
double bfmbase<Float>::inner_real(Fermion_t x_t,Fermion_t y_t)
{
  int nspinco = 12;

  Float *x = (Float *) x_t;
  Float *xx;

  Float *y = (Float *) y_t;
  Float *yy;

  double nrm_re = 0.0;

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=throff;s<throff+thrlen;s++){

    xx = &x[s*nspinco*2];
    yy = &y[s*nspinco*2];

    for(int spinco=0;spinco<nspinco;spinco++){

      int idx_re = 0 + spinco*2;
      int idx_im = 1 + spinco*2;

      nrm_re = nrm_re + xx[idx_re] *yy[idx_re] + xx[idx_im] *yy[idx_im];

  }
  }

  double mydot = nrm_re;
  thread_sum(nrm_re,me);

  if ( reproduce ) { 
    repro(mydot,me);
    if ( reproduce_checksum ) { 
      uint64_t sum=checksum(x_t);
      if ( me == 0 ) repro(sum);
    }
    thread_barrier();
    if ( me == 0 ) repro(nrm_re,nthread);
  }

  comm_gsum(nrm_re);

  return nrm_re;
}
#endif

template <class Float>
void bfmbase<Float>::fill(Fermion_t x_t, double a)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  Float nrm = 0.0;

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=0;s<thrlen;s++){

    xx = &x[(s+throff)*nspinco*2 ];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + spinco*2;
      xx[idx] = a ;
    }}

  }  
  
  thread_barrier();
}

template <class Float>
void bfmbase<Float>::master_fill(Fermion_t x_t, double a)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  Float nrm = 0.0;

  int throff=0;
  int thrlen=cbLs*simd_cbvol*nsimd;
  int me=0;

  for(int s=0;s<thrlen;s++){
    xx = &x[(s+throff)*nspinco*2 ];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + spinco*2;
      xx[idx] = a ;
    }}

  }  
  
}

template <class Float>
double bfmbase<Float>::norm(Fermion_t x_t)
{
  int nspinco = 12;

  Float *x = (Float *) x_t;
  Float *xx;
  double nrm = 0.0;


  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=throff;s<throff+thrlen;s++){

    xx = &x[s*nspinco*2];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + spinco*2;
      nrm = nrm+ xx[idx]*xx[idx];
    }}
  }

  double mydot = nrm;
  thread_sum(nrm,me);

  if ( reproduce ) { 
    repro(mydot,me);
    if ( reproduce_checksum ) { 
      uint64_t sum=checksum(x_t);
      if ( me == 0 ) repro(sum);
    }
    thread_barrier();
    if ( me == 0 ) repro(nrm,nthread);
  }

  comm_gsum(nrm);

  return nrm;
}

template <class Float>
void bfmbase<Float>::conj(Fermion_t x_t)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  Float nrm = 0.0;


  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=0;s<thrlen;s++){

    xx = &x[(s+throff)*nspinco*2 ];

    for(int spinco=0;spinco<nspinco;spinco++){

      int idx_re = 0 + spinco*2;
      int idx_im = 1 + spinco*2;

      xx[idx_re] = xx[idx_re];
      xx[idx_im] = -xx[idx_im];
    
   }

  }  
  
  thread_barrier();
}

template <class Float>
uint64_t bfmbase<Float>::checksum(Fermion_t x_t)
{
  int nspinco = 12;

  Float *x = (Float *) x_t;
  Float *xx;
  union { double d ; uint64_t u ; } conv;
        
  uint64_t csum = 0;

  int me,thrlen,throff;
  thread_work(cbLs*simd_cbvol*nsimd,me,thrlen,throff);

  for(int s=throff;s<throff+thrlen;s++){

    xx = (Float *)&x[s*nspinco*2];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + spinco*2;
      conv.d = xx[idx];
      csum = csum^conv.u;
    }}
  }

  thread_csum(csum,me);

  return csum;
}


template <class Float>
uint64_t bfmbase<Float>::checksumGauge(void)
{
  int ncoco = 9;
  // 8 links per site      => vol * 8
  // 2 sites on inner most => /2
  // dag+undag             => /2
  int vol = node_latt[0]*node_latt[1]*node_latt[2]*node_latt[3]*2;

  Float *xx;
  union { double d ; uint64_t u ; } conv;
        
  uint64_t csum = 0;
  uint64_t csumdag = 0;

  int me,thrlen,throff;
  thread_work(vol,me,thrlen,throff);

  for(int s=throff;s<throff+thrlen;s++){

    xx = (Float *)&u[s*ncoco*2*2*2];
    uint64_t gsum = 0;
    uint64_t gsumdag= 0;
    for(int coco=0;coco<ncoco;coco++){
    for(int reim=0;reim<4;reim++){
      int idx = reim + coco*4;
      conv.d = xx[idx];
      gsum = gsum^conv.u;
      conv.d = xx[idx+36];
      gsumdag = gsumdag^conv.u;
    }}
    csum = csum ^ gsum;
    csumdag = csumdag ^ gsumdag;
  }
  

  thread_csum(csum,me);
  thread_csum(csumdag,me);

  if ( csumdag != csum ){
    printf("Bad dag checksum comparison %16.16lx %16.16lx %lx\n",csum,csumdag, this);
  }

  return csum;
}

template <class Float>
double bfmbase<Float>::norm(Fermion_t x_t,int s)
{
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *xx;
  int me,thrlen,throff;
  thread_work(simd_cbvol,me,thrlen,throff);
  double nrm=0;
  for(int site=0;site<thrlen;site++){

    // Nsimd x 24 x Ls x 4dSite ordering
    xx = &x[((site+throff)*this->cbLs +s)*nspinco*2*this->nsimd];

    for(int idx=0;idx<2*nspinco*this->nsimd;idx++){
      nrm+=xx[idx]*xx[idx];
    }

  }  
  thread_barrier();
  thread_sum(nrm,me);
  comm_gsum(nrm);
  thread_barrier();
  return nrm;

}
template <class Float>
void bfmbase<Float>::axpby_ssp_internal(Fermion_t out,
					double al,double ah,Fermion_t x_t,
					double bl,double bh,Fermion_t y_t,
					int sxo,int sy)
{

#if 1
  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *y = (Float *) y_t;
  Float *o = (Float *) out;

  // write_vaxpy 2 way unrolled operating on 3 colors * complex * nsimd
  double aa_b[4*4+4];
  double *aa = ALIGNIT(aa_b);

  integer args[6];

  int me,thrlen,throff;
  thread_work(simd_cbvol,me,thrlen,throff);

  int ssxo= sxo*24*nsimd;
  int ssy = sy *24*nsimd;

  Float *xx = ((Float * )x) + throff*24*nsimd*cbLs +ssxo;
  Float *yy = ((Float * )y) + throff*24*nsimd*cbLs +ssy;
  Float *oo = ((Float * )o) + throff*24*nsimd*cbLs +ssxo;

  aa[0] = aa[1] = aa[2] = aa[3] = al;
  aa[4] = aa[5] = aa[6] = aa[7] = ah;
  aa[8] = aa[9] = aa[10]= aa[11]= bl;
  aa[12]= aa[13]= aa[14]= aa[15]= bh;

  args[0]=(integer)oo;
  args[1]=(integer)xx;
  args[2]=(integer)yy;
  args[3]=thrlen;
  args[4]=cbLs*nsimd*24*sizeof(Float);
  args[5]=(integer)aa;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpby_ssp((integer *)args);
  } else { 
    vmx_vaxpby_ssp_s((integer *)args);
  }

  thread_barrier();

#else

  int nspinco = 12;
  Float *x = (Float *) x_t;
  Float *y = (Float *) y_t;
  Float *o = (Float *) out;
  Float *xx;
  Float *yy;
  Float *oo;

  int me,thrlen,throff;
  thread_work(simd_cbvol,me,thrlen,throff);

  int skip = 12*this->nsimd;

  for(int site=0;site<thrlen;site++){

    // Nsimd x 24 x Ls x 4dSite ordering
    xx = &x[((site+throff)*this->cbLs +sxo)*nspinco*2*this->nsimd];
    oo = &o[((site+throff)*this->cbLs +sxo)*nspinco*2*this->nsimd];
    yy = &y[((site+throff)*this->cbLs +sy )*nspinco*2*this->nsimd];

    for(int idx=0;idx<2*6*this->nsimd;idx++){
      oo[idx]      = al*xx[idx]+bl*yy[idx];
      oo[idx+skip] = ah*xx[idx+skip]+bh*yy[idx+skip];
    }

  }  

  thread_barrier();
#endif
}
template <class Float>
void bfmbase<Float>::axpby_ssp_proj   (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy,int psign)
{

  if ( psign==1 )          axpby_ssp_internal(out,a,a,x,b,0,y,sxo,sy);
  else if ( psign==-1 )    axpby_ssp_internal(out,a,a,x,0,b,y,sxo,sy);
  else { 
    printf("axpby_ssp_proj bad psign\n");
    exit(-1);
  }

}

template <class Float>
void bfmbase<Float>::ag5xpby_ssp_proj   (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy,int psign)
{
  if ( psign==1 )          axpby_ssp_internal(out,a,-a,x,b,0,y,sxo,sy);
  else if ( psign==-1 )    axpby_ssp_internal(out,a,-a,x,0,b,y,sxo,sy);
  else { 
    printf("ag5xpby_ssp_proj bad psign\n");
    exit(-1);
  }
}

template <class Float>
void bfmbase<Float>::axpby_ssp   (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy)
{
  axpby_ssp_internal(out,a,a,x,b,b,y,sxo,sy);
}
template <class Float>
void bfmbase<Float>::ag5xpby_ssp   (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy)
{
  axpby_ssp_internal(out,a,-a,x,b,b,y,sxo,sy);
}
template <class Float>
void bfmbase<Float>::ag5xpbg5y_ssp (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy)
{
  axpby_ssp_internal(out,a,-a,x,b,-b,y,sxo,sy);
}

template <class Float>
void bfmbase<Float>::chiralproj_axpby(Fermion_t y, Fermion_t x,double a, int sign, int ls_slices)
{
  Float *xx;
  Float *yy;
  double aa_b[8];
  double *aa=ALIGNIT(aa_b);

  int nspinco = 12;

  int me,thrlen,throff;
  thread_work(ls_slices*simd_cbvol,me,thrlen,throff);


  /*
   * 1+g5 -> 1 0
   *         0 0
   * Offset => 0 if sign == -1 
   * 
   * 1-g5 -> 0 0
   *         0 1
   * Offset => 12 if sign == -1 
   *   
   */      
  int offset = throff * 24 *nsimd;

  if ( sign == 1 ) { 
    xx = (Float * )x + offset;
    yy = (Float * )y + offset;
  } else if ( sign == -1 ) { 
    xx = ((Float * )x) + nspinco*nsimd + offset;
    yy = ((Float * )y) + nspinco*nsimd + offset;
  } else { 
    printf("oops - bad sign in chiralproj_axpby\n");
    exit(0);
  }
  
  aa[0] = aa[1] = aa[2] = aa[3] = a;
  //Overwrites yy <- yy + a (1+-g5) xx
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_1pG5vaxpy((double *)yy,(double *)aa,(double *)xx,thrlen);
  } else { 
    vmx_1pG5vaxpy_s((float *)yy,(double *)aa,(float *)xx,thrlen);
  }

  thread_barrier();
}


template <class Float>
void bfmbase<Float>::four_to_five(Fermion_t input[2], Fermion_t result5d[2]){
  printf("four_to_five not implmementd\n");
  exit(0);
}
template <class Float>
void bfmbase<Float>::five_to_four(Fermion_t input5d[2], Fermion_t result[2], int sgn){
  printf("five_to_four not implmementd\n");
  exit(0);
}


#endif
