#include "bfm.h"
#include "bfm_vmx.h"
#include "bfm_mprec.h"
#include "bfm_linalg.h"
#include "bfm_cg.h"
#include "bfm_qdp.h"
#include "bfm_import_export.h"


#include <sys/time.h>

static int bfm_num_inversions;
static struct timeval bfm_FirstInversion;
static struct timeval bfm_InversionTimeToDate;
static struct timeval bfm_InversionEntered;

template <class Float>
void bfmbase<Float>::InverterEnter(void)
{
  if ( bfm_num_inversions == 0 ) {
    gettimeofday(&bfm_FirstInversion,NULL);
  }
  bfm_num_inversions++;
  gettimeofday(&bfm_InversionEntered,NULL);
}
template <class Float>
void bfmbase<Float>::InverterExit (void)
{
  struct timeval now;
  struct timeval elapsed;
  gettimeofday(&now,NULL);
  timersub(&now,&bfm_InversionEntered,&elapsed);
  timeradd(&elapsed,&bfm_InversionTimeToDate,&bfm_InversionTimeToDate);

  double this_inversion = elapsed.tv_sec+1.0e-6*elapsed.tv_usec;
  printf("Inversion :            %.6f s \n",this_inversion);
  double all_inversions = bfm_InversionTimeToDate.tv_sec+1.0e-6*bfm_InversionTimeToDate.tv_usec;
  printf("Inversion time to date %.6f s \n",all_inversions);
  
  timersub(&now,&bfm_FirstInversion,&elapsed);
  double wall_clock =  elapsed.tv_sec+1.0e-6*elapsed.tv_usec;
  printf("Elapsed   time to date %.6f s \n",wall_clock);
  printf("Percentage time in inverter %.3f\n",100.0*all_inversions/wall_clock);

}


template <class Float>
void bfmbase<Float>::comm_gsum(double &val) { return ; };

template <class Float>
bool bfmbase<Float>::isBoss() {return 1;};

// Default values
double bfmarg::mobius_scale = 1.7;
int bfmarg::threads = 1;
int bfmarg::reproduce = 0;
int bfmarg::reproduce_checksum = 0;
int bfmarg::reproduce_master_check= 0;
int bfmarg::verbose= 3;
int bfmarg::CGdiagonalMee= 0;
int bfmarg::part_frac_chroma_convention = 1;
char *bfmarg::watchfile=NULL;
int bfmarg::onepluskappanorm;

template <class Float>
int bfmbase<Float>::CheckStopWatchFile(void)
{
  if ( watchfile ) {
    FILE * fp = fopen(watchfile,"r");
    if ( fp ) {
      fclose(fp);
      printf("Found watchfile %s\n",watchfile);
      return 1;
    }
  }
  return 0;
}

#define MAX_ENTRIES (50000)

#ifdef BFM_REPRO
template <class Float>
void bfmbase<Float>::repro_control(int state) { 

  int me = thread_barrier();
  repro_state = state; 

  thread_barrier();

  if ( repro_state == ReproIdle ) return;

  repro_entry = 0;
  csum_entry = 0;

  if ( (repro_state == ReproRecord) && (me == 0) ){
    repro_log.resize(nthread+1);
    for(int tid=0;tid<nthread+1;tid++){
      repro_log[tid].resize(MAX_ENTRIES);
    }
    csum_log.resize(MAX_ENTRIES);
  }
  thread_barrier();
};

template <class Float>
void bfmbase<Float>::repro(uint64_t csum) {

  if ( csum_entry >= MAX_ENTRIES-1 )  return;

  if ( repro_state == ReproRecord ) csum_log[csum_entry] = csum;

  if ( repro_state == ReproCheck  ) {
    uint64_t pop = csum_log[csum_entry];
    if( csum != pop ) { 
    
      fprintf(stderr,"Repro mismatch %lx e=%d csum : %16.16lx %16.16lx\n",
	      (uint64_t)this,csum_entry,csum,pop);
      fprintf(stderr,"return addresses %lx %lx %lx\n",
	     __builtin_return_address(0),
	     __builtin_return_address(1),
	     __builtin_return_address(2)
	      );
      fflush(stderr);
      abort();
    }
  }

  csum_entry++;

}
template <class Float>
void bfmbase<Float>::repro(double d,int who) {

  if ( repro_entry >= MAX_ENTRIES-1 ) return;

  if ( (who < 0) || (who > nthread) ) { 
    printf("repro: bad who %d\n",who);
    exit(-1);
  }

  if ( repro_state == ReproRecord ) repro_log[who][repro_entry] = d;
  if ( repro_state == ReproCheck  ) {
    double pop = repro_log[who][repro_entry];
    if( d != pop ) { 
      fprintf(stderr,"Repro mismatch %lx e=%d : %16.8le %16.8le %16.16lx %16.16lx lane %d\n",
	      (uint64_t)this,repro_entry,d,pop,*((uint64_t*)&d),*((uint64_t *)&pop),who);

      fprintf(stderr,"return addresses %lx %lx %lx\n",
	     __builtin_return_address(0),
	     __builtin_return_address(1),
	     __builtin_return_address(2)
	      );
      fflush(stderr);
      abort();
    } 
  }
  if ( who ==nthread ) {
    repro_entry++;
  }
};
#endif

template <class Float>
int bfmbase<Float>::doubleStored(void) { 
  return 1;
};

template class bfmbase<double>;
template class bfmbase<float>;
