#include <bfm.h>
#include <stdio.h>
#include <stdlib.h>

#include <bfmcommqmp.h>
#include <qmp.h>

extern int time_report;


/*
 * QMP comms implementation
 * If ( local_comm[mu]== 1 ) these routines do not get called
 * If ( local_comm[mu]== 0 ) these routines get called but loop back
 *
 * For now a brain dead implementation with no spin projection.
 */

#define recv_multi_handle multi_handle[0]
#define send_multi_handle multi_handle[1]

template <class Float> 
void bfmcommQMP<Float>::recv_end(void)
{
  if ( num_op>0 ) {
    QMP_free_msghandle(recv_multi_handle);
  }
  for(int mu=0; mu < ND; mu++) { 
    if (!this->local_comm[mu]) { 
      QMP_free_msgmem(recv_ops_msgmem_t[2*mu]);
      QMP_free_msgmem(recv_ops_msgmem_t[2*mu+1]);
      if ( this->simd_grid[mu]>1 ){
	bfm_free(simd_rbuf[2*mu]);
	bfm_free(simd_rbuf[2*mu+1]);
      }
    }
  }
}
template <class Float> 
void bfmcommQMP<Float>::recv_init(Fermion_t psi_t)
{
  num_op = 0;
  Float *psi = (Float *) psi_t;  
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++ ) {

      int idx=2*mu+pm; 
      
      this->recvbufs[idx] = &psi[this->comm_offset[pm][mu]*SPINOR_SIZE*this->simd()*this->cbLs];

      if( !this->local_comm[mu] ) {

	
	// receive buffers
	// For a SIMD direction should reduce SIMD() factor....
	int simd_factor = this->simd();
	simd_factor = simd_factor/this->simd_grid[mu];

	int words  = this->simd_nbound[mu] * HALF_SPINOR_SIZE * simd_factor * this->cbLs;
	int delta;

	if ( pm == Plus ) delta = 1;
	else delta = -1;

	Float *rbuf = this->recvbufs[idx];
	if ( this->simd_grid[mu] > 1 ) {
	  rbuf       = simd_rbuf[idx] = (Float *)bfm_alloc(words*sizeof(Float));
	}
	recv_ops_msgmem_t[idx]  = QMP_declare_msgmem((void *)rbuf,words*sizeof(Float));
	recv_handles[num_op++]  = QMP_declare_receive_relative(recv_ops_msgmem_t[idx], mu, delta, 0);

      } // if ( !local_comm )
    }      
    } // for(mu=... )
    
    if( num_op > 0 ) { 
      recv_multi_handle = QMP_declare_multiple(recv_handles, num_op);
    }

  return;

}
template <class Float> 
void bfmcommQMP<Float>::comm_init(void)
{
                                                                            
  num_op = 0;
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++ ) {

      int words, idx, alloc_words;
      Float * ptr;

      // Allocate send buffers
      idx=2*mu+pm; 

      int simd_factor = this->simd();
      simd_factor = simd_factor/this->simd_grid[mu];

      words = this->simd_nbound[mu] * HALF_SPINOR_SIZE * simd_factor * this->cbLs;
      alloc_words = this->simd_nbound[mu] * HALF_SPINOR_SIZE * this->simd() * this->cbLs;
      ptr= this->sendbufs[idx] = (Float *)bfm_alloc(alloc_words*sizeof(Float)); 
      if( !this->local_comm[mu] ) {

	// Declare send & msgmem
	int delta;
	if ( pm == Plus ) delta = 1;
	else delta = -1;

	// We pull the "Plus" face_table and send in the Minus direction
	// This reflects an oddity of the 
	send_ops_msgmem_t[idx] = QMP_declare_msgmem((void *)ptr,words*sizeof(Float));
	send_handles[num_op++] = QMP_declare_send_relative(send_ops_msgmem_t[idx], mu, -delta, 0);

      } 
    }      
    } // for(mu=... )
    
    if( num_op > 0 ) { 
      send_multi_handle = QMP_declare_multiple(send_handles, num_op);
    }


  return;
}

template <class Float> 
void bfmcommQMP<Float>::comm_end(void)
{
  /*-----------------------------------------------------------------------* 
   * I need to free the multiple msghandles                                *
   * The individual handles (tmp_handles) used should have been freed      *
   * when the multiple was created and according to the spec I don't need  *
   * to free them                                                          *
   *---------------------------------------------------------------------- */
  if ( num_op>0 ) QMP_free_msghandle(send_multi_handle);
  /*---------------------------------------------------------------------*
   * I need to free the associated msgmem_t-s here.                      *
   *---------------------------------------------------------------------*/
  for(int mu=0; mu < ND; mu++) { 
    bfm_free(this->sendbufs[2*mu]);
    bfm_free(this->sendbufs[2*mu+1]);
    if (!this->local_comm[mu]) { 
      QMP_free_msgmem(send_ops_msgmem_t[2*mu]);
      QMP_free_msgmem(send_ops_msgmem_t[2*mu+1]);
    }
  }

  return;
}

template <class Float> 
void bfmcommQMP<Float>::comm  (int result_cb, Fermion_t psi,int dag)  
{
  comm_start(result_cb,psi,dag);
  comm_complete(result_cb,psi);
}


template <class Float> 
void bfmcommQMP<Float>::comm_start  (int result_cb, Fermion_t psi,int dag)  
{
  // gather the faces. Routines here are threaded.
  // All threads cooperate in gathering.
  

  //  if ( this->isBoss() && (me == 0)) {
  //    printf("comm_start checking heap\n"); fflush(stdout);
  //    mcheck_check_all();
  //    printf("comm_start mcheck_all");fflush(stdout);
  //  }

  int me = this->thread_barrier();
  if ( me == 0 ) {
    recv_init((Fermion_t)psi);
  }
  this->thread_barrier();

  gather (result_cb,psi,dag);

  this->thread_barrier();
  if ( me == 0 ) {
    comm_qmp_start(psi);
  }
  this->thread_barrier();

  return;
}
template <class Float> 
void bfmcommQMP<Float>::comm_qmp_start(Fermion_t psi) 
{
    if ( num_op > 0 ) {
      QMP_start(recv_multi_handle);
      QMP_start(send_multi_handle);
    }
}
template <class Float> 
void bfmcommQMP<Float>::comm_qmp_complete(void) 
{
  if ( num_op > 0 ) { 
    
    if( QMP_wait(recv_multi_handle) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }
    
    if( QMP_wait(send_multi_handle) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }

  }
}

template <class Float> 
void bfmcommQMP<Float>::comm_merge  (void)  
{
#undef TEST
#ifdef TEST
  int me=this->thread_barrier();
  
  for(int pm=0;pm<2;pm++){
    int idx= 6+pm;
    Float *rbuf       = (Float *)((integer)this->simd_rbuf[idx]);
    if( !this->local_comm[3] ) {
      for(int site=0;site<6;site++){
	printf("rbuf_qmp %lx %4.4d %4.4d\n",(integer)rbuf,	       (int)rbuf[0],(int)rbuf[1]	       );
	rbuf=rbuf+12;
      }
    }
  }
#endif 

  for (int mu=0;mu<4;mu++) { 
    for(int pm=0;pm<2;pm++){
      int dir=pm+2*mu;
      if ( !this->local_comm[mu] && this->simd_grid[mu]>1 ) {
#ifdef TEST
	printf("===================\n");
	printf("Merge\n");
	printf("===================\n");
#endif
	int length = this->simd_nbound[mu]*this->cbLs;
	int skip   = length*12*sizeof(Float); // NB no nsimd
	Float *off_node = simd_rbuf[dir];
	Float * on_node = (Float *)((integer)this->sendbufs[dir]   + skip);

#ifdef TEST
	Float * rbp = this->recvbufs[dir];
	Float * onp  = on_node;
	Float * offp = off_node;
	for(int s=0;s<length;s++){
	  printf("dir %d Merging off : %4.4d on: %4.4d\n",dir,(int) *offp,(int) *onp);
	  for(int spinco=0;spinco<6;spinco++){
	    onp++;
	    onp++;
	    offp++;
	    offp++;
	  }
	}
#endif
	if( pm==0) {
	  this->merge(this->recvbufs[dir],off_node,on_node,length);
	} else { 
	  this->merge(this->recvbufs[dir],on_node,off_node,length);
	}
      }
    }
  }
#ifdef TEST
  this->thread_barrier();

  if ( this->isBoss()|| 1 ) {
  if (me ==0){
  for(int mu=0;mu<ND;mu++){
    for(int pm=0;pm<2;pm++){
      
      if ( this->simd_grid[mu] > 1 ) {
      printf("====================\n");
      printf("mu %d pm %d %d elements local comm %d\n",mu,pm, this->simd_nbound[mu]*this->cbLs*12,this->local_comm[mu]);
      printf("====================\n");
      int dir=pm+2*mu;

      int length = this->simd_nbound[mu]*this->cbLs;
      int skip   = length*12*sizeof(Float); // NB no nsimd
      Float *rbuf      =this->recvbufs[dir];
      Float *sbuf_off  =this->sendbufs[dir];
      Float *sbuf_on   = (Float *)((integer)this->sendbufs[dir]   + skip);
      printf("rbuf %lx sbuf_off %lx sbuf_on %lx\n",rbuf,sbuf_off,sbuf_on);
      for(int site=0;site<this->simd_nbound[mu]*this->cbLs;site++){
	//      for(int site=0;site<6;site++){
	printf("rbuf %4.4d %4.4d %4.4d %4.4d, sbuf_off %4.4d %4.4d sbuf_on %4.4d %4.4d\n",
	       (int)rbuf[0],(int)rbuf[1],(int)rbuf[2],(int)rbuf[3],(int)sbuf_off[0],(int)sbuf_off[1],
	       (int)sbuf_on[0],(int)sbuf_on[1]
	       );
	rbuf=rbuf+4*6;
	sbuf_off=sbuf_off+2*6;
	sbuf_on =sbuf_on+2*6;
      }
      printf("============\n");
      fflush(stdout);
      }
    }
  }}
  }
  this->thread_barrier();
#endif
}

template <class Float> 
void bfmcommQMP<Float>::comm_complete  (int result_cb, Fermion_t psi)  
{
  uint64_t t1 = GetTimeBase();
  int me=this->thread_barrier();
  if ( me == 0) { 
    comm_qmp_complete() ;
  }
  this->thread_barrier();
  uint64_t t2 = GetTimeBase();

  this->comm_merge();
  uint64_t t3 = GetTimeBase();

  this->thread_barrier();

  if ( me == 0) {
    this->recv_end();
  }
  uint64_t t4 = GetTimeBase();

  if ( this->iter == this->time_report_iter+1 ) {
    if ( this->isBoss() && (me ==0) ){
      printf("qmp_complete %ld cyc\n",t2-t1);
      printf("merge        %ld cyc\n",t3-t2);
      printf("recv_end     %ld cyc\n",t4-t3);
    }
  }
  this->thread_barrier();
  
  return;
}

template <class Float> 
bool bfmcommQMP<Float>::isBoss(void)
{
  return QMP_is_primary_node() ;
}

template <class Float> 
void bfmcommQMP<Float>::gather(int result_cb, Fermion_t psi,int dag)
{
  int nspinco = 12;
  int cb = 1-result_cb;

/**********************************************************
 * Gather the faces into the communications buffers
 **********************************************************
 *
 * vmx_gather_proj     blasts this into the recv buffer "densely", i.e. 4spin->2spin
 *                     for communication (half data).
 * vmx_gather_permute  Interchanges the simd lanes
 * vmx_gather          4 spinor -> 4 spinor gather
 * vmx_gather          4 spinor -> 4 spinor gather

 * -> Gathers buffers at comm offset[pm][mu] from
 *    face_table[result_cb][pm][mu]
 **********************************************************
 */
  int me = this->thread_barrier();

  for(int mu=0;mu<ND;mu++){
    for(int pm=0;pm<2;pm++){

      int permute=0;
      int extract=0;
      int ftcb =0;
      int dir=pm+2*mu;
      int sgn =pm;
      Float *comm_buf = this->sendbufs[dir];

      if ( dag ) sgn = 1-pm;

      if ( this->precon_5d ) ftcb = 0;
      else ftcb = result_cb;

      if ( (this->simd_grid[mu]>1) && (this->local_comm[mu]) ){

	/******************************************************
	 * Direction with SIMD interleave is special
	 * Gather into recvbuf with permute set if local.
	 ******************************************************/
        permute = 1;
	comm_buf = this->recvbufs[dir];

      }

      if ( (this->simd_grid[mu]>1) && (!this->local_comm[mu]) ){
	/******************************************************
	 * Direction with SIMD interleave is special
	 * Gather into sendbuf extracting subset of SIMD vec 
	 ******************************************************/
#ifdef TEST
	printf("Experimental code do not trust this yet : non-local in SIMD direction\n");
#endif
	// Here we must take array : (AE BF CG DH) (IM JN KO LP)
	// Pull AE in, A-> sendbuf
	//
	// After receive of "I" from next node into the receive buf.
	// Should receive into high part to enable in place merge.
	// This may come later, for now can do it poorly prefetched descending order
        extract = 1;
      }

      int do_comm = 0;
      if ( !this->local_comm[mu]   ) do_comm=1;
      if ( this->simd_grid[mu] > 1 ) do_comm=1;
      if ( this->simd_grid[mu] > 2 ) {
	printf("simd_grid[%d] > 2 not supported\n",mu);
	exit(0); // Not supported yet
      }
      if ( this->simd()        > 2 ) {
	printf("simd > 2 not supported\n");
	exit(0); // Not supported yet
      }


      if (do_comm) {

	int thrvol,throff;
	this->thread_work_nobarrier(this->simd_nbound[mu],me,thrvol,throff);

	integer args[9];
	args[0] = (integer) &this->face_table[ftcb][pm][mu][throff];
	args[1] = (integer) psi;
	args[2] = (integer) throff;
	args[3] = (integer) comm_buf; 
	args[4] = (integer) thrvol;
	args[5] = (integer) this->cbLs;
	args[6] = sgn+2*mu;
	args[7] = (integer) this->complex_i_simd;

	integer arg_p = (integer) args;

	if ( permute ) { 

#ifdef TEST
	  if ( me==0 ) {
	    printf("mu %d pm %d -- doing permute!\n",mu,pm);
	  }
#endif
	  if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj_perm(arg_p);
	  else                                    vmx_gather_proj_perm_s(arg_p);

	} else if ( extract ) { 
#ifdef TEST
	  if ( me==0 ) {
	    printf("TEST Experimentally filling send buf with SIMD extract code\n");
	  }
#endif
	  args[3] = (integer) this->sendbufs[dir]; // Off node
	  args[8] = (integer) this->sendbufs[dir]  // On node
	    + this->simd_nbound[mu]*this->cbLs*12*sizeof(Float); // NB no nsimd
	  if ( pm == 0 ) {
	    integer tmp =args[3];
	    args[3]=args[8];
	    args[8]=tmp;
	  }

	  if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj_extr_l(arg_p);
	  else                                    vmx_gather_proj_extr_l_s(arg_p);
	  
	} else { 
	  if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj(arg_p);
	  else                                    vmx_gather_proj_s(arg_p);
	}  


      }
    }
  }

  this->thread_barrier();

#ifdef TEST
  if ( this->isBoss() ) {
  int length = this->simd_nbound[3]*this->cbLs;
  int skip   = length*12*sizeof(Float); // NB no nsimd
  if (me ==0){
  for(int mu=0;mu<ND;mu++){
    for(int pm=0;pm<2;pm++){
      
      if ( this->simd_grid[mu] > 1 ) {
      printf("====================\n");
      printf("mu %d pm %d %d elements local comm %d\n",mu,pm, this->simd_nbound[mu]*this->cbLs*12,this->local_comm[mu]);
      printf("====================\n");
      int dir=pm+2*mu;

      Float *sbuf_off  =this->sendbufs[dir];
      Float *sbuf_on   = (Float *)((integer)this->sendbufs[dir]   + skip);

      for(int site=0;site<this->simd_nbound[mu]*this->cbLs;site++){
	//      for(int site=0;site<6;site++){
	printf("gather sbuf_off %4.4d %4.4d sbuf_on %4.4d %4.4d\n",
	       (int)sbuf_off[0],(int)sbuf_off[1],
	       (int)sbuf_on[0],(int)sbuf_on[1]
	       );
	sbuf_off=sbuf_off+2*6;
	sbuf_on =sbuf_on+2*6;
      }
      printf("============\n");
      fflush(stdout);
      }
    }
  }}
  }
#endif
  this->thread_barrier();
}


template <class Float> 
void bfmcommQMP<Float>::comm_gsum(double &val) 
{
  int me = this->thread_barrier();
  double myval=0;
  if ( me == 0 ) { 
    QMP_sum_double_array(&val,1);
    myval = val;
  }
  this->thread_sum(myval,me);
  val = myval;
}

template class bfmcommQMP <float>;
template class bfmcommQMP <double>;

